#
# This piece of code is written by
#    Jianing Yang <jianingy.yang@gmail.com>
# with love and passion!
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#                             17 Mar, 2016
#
from __future__ import absolute_import
from . import service
from .utils import fatal, out, header_out
from humanize import naturalsize, naturaltime
from tornado.gen import coroutine
from tornado.log import enable_pretty_logging
from tornado.ioloop import IOLoop
from tornado.options import options, define

import click
import functools
import six
import traceback


def pretty_exception(fn):

    @functools.wraps(fn)
    def wrap(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            if options.debug:
                traceback.print_exc(e)
            fatal(e.message)

    return wrap


def cmp_host(x, y):
    if x > y:
        return 1
    return -1


def _val(d, path, transformer=lambda x: unicode(x)):
    current_value = d
    try:
        for node in path.split('.'):
            if node.isdigit():
                current_value = current_value[int(node)]
            else:
                current_value = current_value[node]
    except KeyError:
        return 'n/a'
    else:
        return transformer(current_value)


@click.group()
@click.option('--debug', flag_value=True, default=False,
              help='enable debug mode')
@click.option('--timeout', default=5, help='timeout')
@click.option('--group-dir', default='groups',
              help='directory of group definition files')
def wac(group_dir, debug, timeout):
    define('debug', default=debug)
    define('timeout', default=timeout)
    define('group-dir', default=group_dir)
    if debug:
        options.logging = 'debug'

    enable_pretty_logging()


@wac.command(name='client-list',
             help='list all connected devices')
@click.argument('host', type=str, nargs=-1)
def client_list(host):

    hosts = service.expand_host(host[0])

    @coroutine
    def _run():
        aps = yield service.create_multiple_ap(hosts)
        clients = yield service.client_list(aps)
        header_out('name, host, ifname, freq, mac_address')
        out('\n'.join(['{name} {host} {ifname} {freq} {mac}'.format(**c)
                       for c in clients]))

    IOLoop.instance().run_sync(_run)


@wac.command(name='client-ban', help='ban a client by mac address')
@click.option('--wlan-id', type=int, help='wlan id')
@click.option('--mac', type=str, help='mac address to ban')
@click.argument('host', type=str, nargs=-1)
@pretty_exception
def client_ban(wlan_id, mac, host):

    hosts = service.expand_host(host[0])

    @coroutine
    def _run():
        aps = yield service.create_multiple_ap(hosts)
        header_out('host, wlan_id, reason')
        for ap in aps.values():
            status = yield service.client_ban(ap, wlan_id, mac)
            out('%s, %s, %s' % (ap.addr, wlan_id, status[1]))

    IOLoop.instance().run_sync(_run)


@wac.command(name='client-permit', help="restore a client's access")
@click.option('--wlan-id', type=int, help='wlan id')
@click.option('--mac', type=str, help='mac address to permit')
@click.argument('host', type=str, nargs=-1)
@pretty_exception
def client_permit(wlan_id, mac, host):

    hosts = service.expand_host(host[0])

    @coroutine
    def _run():
        aps = yield service.create_multiple_ap(hosts)
        header_out('host, wlan_id, reason')
        for ap in aps.values():
            status = yield service.client_permit(ap, wlan_id, mac)
            out('%s, %s, %s' % (ap.addr, wlan_id, status[1]))

    IOLoop.instance().run_sync(_run)


@wac.command(name='ap-list', help='show all openwrt devices status')
@click.argument('host', type=str, nargs=-1)
def ap_list(host):

    hosts = service.expand_host(host[0])

    def _calc_load(x):
        return x / 65535

    @coroutine
    def _run():
        rows = []
        aps = yield service.create_multiple_ap(hosts)
        details = yield service.ap_list(aps)
        header_out('name, host, #clients, loadavg, mem, uptime')
        for ap in details:
            row = []
            row.append(_val(ap, 'board.hostname'))
            row.append(_val(ap, 'host'))
            row.append('%s' % (_val(ap, 'num_clients')))
            row.append('%.2f / %.2f / %.2f' % (
                _val(ap, 'system.load.0', _calc_load),
                _val(ap, 'system.load.1', _calc_load),
                _val(ap, 'system.load.2', _calc_load)))
            row.append('%s / %s' %
                       (_val(ap, 'system.memory.free', naturalsize),
                        _val(ap, 'system.memory.total', naturalsize)))
            row.append('%s' % (_val(ap, 'system.uptime', naturaltime)))
            rows.append(', '.join(row))

        out('\n'.join(sorted(rows, cmp_host)))
    IOLoop.instance().run_sync(_run)


@wac.command(name='wlan-edit', help='change wlan settings')
@click.option('--wlan-id', type=int, default=0,
              help='wlan id')
@click.option('--enable', 'state', flag_value='enable', help='disable wlan')
@click.option('--disable', 'state', flag_value='disable', help='disable wlan')
@click.option('--ssid', type=str, help='new ssid to change')
@click.option('--channel', type=int, help='new channel to change')
@click.option('--encryption', type=str,
              help='new password to change. format: "method:secret"')
@click.argument('host', type=str, nargs=-1)
@pretty_exception
def wlan_edit(wlan_id, state, channel, ssid, encryption, host):

    hosts = service.expand_host(host[0])

    @coroutine
    def _run():
        aps = yield service.create_multiple_ap(hosts)
        for ap in aps.values():
            if channel:
                status = yield service.wlan_set_channel(ap, channel, wlan_id)
                out('%s, wlan%d, %s' % (ap.addr, wlan_id, status[1]))
            if ssid:
                status = yield service.wlan_set_ssid(ap, ssid, wlan_id)
                out('%s, wlan%d, %s' % (ap.addr, wlan_id, status[1]))
            if encryption:
                method, _, key = encryption.partition(':')
                status = yield service.wlan_set_encryption(ap, method, key,
                                                           wlan_id)
                out('%s, wlan%d, %s' % (ap.addr, wlan_id, status[1]))
            if state == 'enable':
                status = yield service.wlan_enable(ap, True, wlan_id)
                out('%s, wlan%d, %s' % (ap.addr, wlan_id, status[1]))
            if state == 'disable':
                status = yield service.wlan_enable(ap, False, wlan_id)
                out('%s, wlan%d, %s' % (ap.addr, wlan_id, status[1]))

    IOLoop.instance().run_sync(_run)


@wac.command(name='wlan-list', help='show all wlan interfaces status')
@click.argument('host', type=str, nargs=-1)
def wlan_list(host):

    hosts = service.expand_host(host[0])

    @coroutine
    def _run():
        rows = []
        aps = yield service.create_multiple_ap(hosts)
        details = yield service.wlan_list(aps)
        header_out('name, host, radio, wlan, ssid, enc, '
                   'status, hwmode, htmode, '
                   'txpower, channel')
        for ap in details:
            for radio, status in six.iteritems(ap['wireless']):
                for wlan in status['interfaces']:
                    row = []
                    row.append(_val(ap, 'board.hostname'))
                    row.append(_val(ap, 'host'))
                    row.append('%s' % radio)
                    row.append('%s' % _val(wlan, 'ifname'))
                    row.append('%s' % _val(wlan, 'config.ssid'))
                    row.append('%s' % _val(wlan, 'config.encryption'))
                    row.append(_val(status, 'up',
                                    lambda x: 'up' if x else 'down'))
                    row.append(_val(status, 'config.hwmode'))
                    row.append(_val(status, 'config.htmode'))
                    row.append(_val(status, 'config.txpower'))
                    row.append(_val(status, 'config.channel'))
                    rows.append(', '.join(row))

        out('\n'.join(sorted(rows, cmp_host)))
    IOLoop.instance().run_sync(_run)



#
# This piece of code is written by
#    Jianing Yang <jianingy.yang@gmail.com>
# with love and passion!
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#                             17 Mar, 2016
#
from __future__ import absolute_import
from .exception import GenericError
from json import dumps as json_encode, loads as json_decode
from tornado.gen import coroutine, Return
from tornado.httpclient import AsyncHTTPClient
from tornado.options import options
from uuid import uuid4
import logging

LOG = logging.getLogger('tornado.application')

UBUS_STATUS_OK = 0
UBUS_STATUS_INVALID_COMMAND = 1
UBUS_STATUS_INVALID_ARGUMENT = 2
UBUS_STATUS_METHOD_NOT_FOUND = 3
UBUS_STATUS_NOT_FOUND = 4
UBUS_STATUS_NO_DATA = 5
UBUS_STATUS_PERMISSION_DENIED = 6
UBUS_STATUS_TIMEOUT = 7
UBUS_STATUS_NOT_SUPPORTED = 8
UBUS_STATUS_UNKNOWN_ERROR = 9
UBUS_STATUS_CONNECTION_FAILED = 10


class UBusServerError(GenericError):
    error_format = 'uci server error: %(reason)s'


class UBusClientError(GenericError):
    error_format = 'ERROR[%(code)s]: %(reason)s'


class UBusClient(object):

    def __init__(self, addr, port=80, username='root', password='daling.com'):
        self.addr = addr
        self.port = port
        self.session = '00000000000000000000000000000000'
        self.username = username
        self.password = password
        self.api_root = 'http://{host}:{port}/ubus'.format(host=addr,
                                                           port=port)

    @coroutine
    def _rpc_call(self, method, *args):
        http_client = AsyncHTTPClient()
        request_id = str(uuid4())
        payload = json_encode({'jsonrpc': '2.0',
                               'id': request_id,
                               'method': method,
                               'params': args})
        resp = yield http_client.fetch(self.api_root,
                                       connect_timeout=options.timeout,
                                       request_timeout=options.timeout,
                                       method='POST',
                                       body=payload)
        if resp.code != 200:
            raise UBusServerError(reason='server returns %s' % resp.code)

        try:
            LOG.debug('rpc returns: %s' % resp.body)
            data = json_decode(resp.body)
        except ValueError:
            raise UBusServerError(reason='server returns malformed json')
        else:
            if data['id'] != request_id:
                UBusServerError(reason='server did not returns'
                                ' correspond request_id')
            if 'result' in data:
                raise Return(data['result'])
            elif 'error' in data:
                error = data['error']
                raise UBusClientError(code=error['code'],
                                      reason=error['message'])
            else:
                UBusServerError(reason='server returns malformed json')

    def _status(self, code):
        return code

    @coroutine
    def call(self, *args):
        params = (self.session,) + args
        result = yield self._rpc_call('call', *params)
        raise Return(result)

    @coroutine
    def list(self, *args):
        params = (self.session,) + args
        result = yield self._rpc_call('list', *params)
        raise Return(result)

    @coroutine
    def login(self, username='root', password='daling.com'):
        cred = dict(username=username, password=password)
        result = yield self.call('session', 'login', cred)
        self.session = result[1]['ubus_rpc_session']
        board_status = yield self.get_board_status()
        if board_status:
            self.kernel = board_status.get('kernel', '')
            self.hostname = board_status.get('hostname', '')
            self.system = board_status.get('system', '')
            self.model = board_status.get('model', '')
            self.release = board_status.get('release', '')

    @coroutine
    def list_clients(self):
        clients = dict()
        hostapds = yield self.list('hostapd.*')
        for hostapd in hostapds.keys():
            clients[hostapd] = (yield self.call(hostapd, 'get_clients', {}))[1]
        raise Return(clients)

    @coroutine
    def del_client(self, wlan_id, mac, ban_time=0, reason=1, deauth=True):
        result = yield self.call('hostapd.wlan%d' % wlan_id,
                                 'del_client',
                                 {'addr': mac,
                                  'reason': reason,
                                  'deauth': deauth,
                                  'ban_time': ban_time})
        raise Return(self._status(result[0]))

    @coroutine
    def get_mac_list(self, wlan_id):
        result = yield self.call('uci', 'get',
                                 {'config': 'wireless',
                                  'section': '@wifi-iface[%d]' % wlan_id,
                                  'option': 'maclist'})
        if result[0] == 0:
            raise Return(result[1]['value'] if len(result) > 1 else [])
        else:
            raise UBusServerError(reason=self._status(result[0]))

    @coroutine
    def set_channel(self, channel, wlan_id):
        result = yield self.call('uci', 'set',
                                 {'config': 'wireless',
                                  'section': 'radio%d' % wlan_id,
                                  'values': {'channel': channel}})
        raise Return(self._status(result[0]))

    @coroutine
    def set_mac_filter_policy(self, policy, wlan_id):
        result = yield self.call('uci', 'set',
                                 {'config': 'wireless',
                                  'section': '@wifi-iface[%d]' % wlan_id,
                                  'values': {'macfilter': policy}})
        raise Return(self._status(result[0]))

    @coroutine
    def set_mac_list(self, mac_list, wlan_id):
        result = yield self.call('uci', 'set',
                                 {'config': 'wireless',
                                  'section': '@wifi-iface[%d]' % wlan_id,
                                  'values': {'maclist': mac_list}})
        raise Return(self._status(result[0]))

    @coroutine
    def set_ssid(self, ssid, wlan_id):
        result = yield self.call('uci', 'set',
                                 {'config': 'wireless',
                                  'section': '@wifi-iface[%d]' % wlan_id,
                                  'values': {'ssid': ssid}})
        raise Return(self._status(result[0]))

    @coroutine
    def set_encryption(self, method, key, wlan_id):
        result = yield self.call('uci', 'set',
                                 {'config': 'wireless',
                                  'section': '@wifi-iface[%d]' % wlan_id,
                                  'values': {'encryption': method,
                                             'key': key}})
        raise Return(self._status(result[0]))

    @coroutine
    def enable_wlan(self, enable, wlan_id):
        result = yield self.call('uci', 'set',
                                 {'config': 'wireless',
                                  'section': '@wifi-iface[%d]' % wlan_id,
                                  'values': {'disabled': 0 if enable else 1}})
        raise Return(self._status(result[0]))

    @coroutine
    def commit(self, config):
        result = yield self.call('uci', 'commit', {'config': config})
        raise Return(self._status(result[0]))

    @coroutine
    def confirm(self, config):
        result = yield self.call('uci', 'confirm', {'config': config})
        raise Return(self._status(result[0]))

    @coroutine
    def get_system_status(self):
        result = yield self.call('system', 'info', {})
        if result[0] != UBUS_STATUS_OK:
            raise Return(dict())
        raise Return(result[1])

    @coroutine
    def get_wireless_status(self):
        result = yield self.call('network.wireless', 'status', {})
        if result[0] != UBUS_STATUS_OK:
            raise Return(dict())
        raise Return(result[1])

    @coroutine
    def get_board_status(self):
        result = yield self.call('system', 'board', {})
        if result[0] != UBUS_STATUS_OK:
            raise Return(dict())
        raise Return(result[1])



# -*- coding: utf-8 -*-
#
# Copyright 2015, Jianing Yang
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# Author: Jianing Yang <jianingy.yang@gmail.com>


import setuptools

setuptools.setup(
    name='openwrt-wac',
    version='0.1.1',
    description=('a simple command line tool for managing '
                 'groups of openwrt based rotuers'),
    author='Jianing Yang',
    author_email='jianingy.yang@gmail.com',
    url='http://github.com/jianingy/openwrt-wac',
    install_requires=['tornado', 'click', 'humanize'],
    packages=['wac'],
    entry_points={
        'console_scripts': [
            'wac=wac.cmd:wac',
            ]
        },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Topic :: Utilities',
    ],
)



#
# This piece of code is written by
#    Jianing Yang <jianingy.yang@gmail.com>
# with love and passion!
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#                             18 Mar, 2016
#
from __future__ import absolute_import
from .exception import GenericError
from .utils import fatal
from .ubus import UBusClient, UBUS_STATUS_OK
from os.path import join as path_join
from tornado.gen import coroutine, Return
from tornado.options import options
import logging
import six

LOG = logging.getLogger('tornado.application')

FREQS_24 = {
    1: 2412, 2: 2417, 3: 2422, 4: 2427, 5: 2432, 6: 2437,
    7: 2442, 8: 2447, 9: 2452, 10: 2457, 11: 2462, 12: 2467,
    13: 2472
}
FREQS_50 = {
    36: 5180, 40: 5200, 44: 5220, 48: 5240, 52: 5260,
    56: 5280, 60: 5300, 64: 5320,
    149: 5745, 153: 5765, 157: 5785, 161: 5805, 165: 5825
}


class InvalidChannelError(GenericError):
    error_format = 'invalid wlan frequency: %(channel)s'


def expand_host(dest):
    if dest.startswith('@'):
        group_name = dest[1:]
        try:
            rows = file(path_join(options.group_dir, group_name)).readlines()
            dests = map(lambda x: x.strip(), rows)
        except IOError:
            fatal('cannot find group %s' % group_name)
            return []
        except Exception:
            raise
        else:
            return dests
    else:
        return [dest]


def raise_on_ubus_error(code, reason='unknown error'):
    if code == UBUS_STATUS_OK:
        return
    raise Return(('False', '[ERROR:%d] %s' % (code, reason)))


@coroutine
def create_multiple_ap(dest):
    aps = dict()
    for dest in dest:
        try:
            ap = UBusClient(dest)
            yield ap.login()
        except Exception as e:
            LOG.warn('unable to log into %s: %s' % (dest, str(e)))
        else:
            aps[dest] = ap
    raise Return(aps)


@coroutine
def client_list(aps):
    client_all = []
    for host, ap in six.iteritems(aps):
        hostapd_clients = yield ap.list_clients()
        for hostapd, retval in six.iteritems(hostapd_clients):
            freq = retval['freq']
            clients = retval['clients']
            ifname = hostapd.replace('hostapd.', '')
            client_all.extend([{'name': ap.hostname,
                                'host': host,
                                'ifname': ifname,
                                'freq': freq,
                                'mac': mac}
                               for mac in clients.keys()])
    raise Return(client_all)


@coroutine
def client_ban(ap, wlan_id, mac):
    result = yield ap.set_mac_filter_policy('deny', wlan_id)
    raise_on_ubus_error(result, 'cannot set policy')
    mac_list = yield ap.get_mac_list(wlan_id)
    mac_list.append(mac)
    result = yield ap.set_mac_list(list(set(mac_list)), wlan_id)
    raise_on_ubus_error(result, 'cannot set maclist')
    result = yield ap.commit('wireless')
    raise_on_ubus_error(result, 'cannot commit configuration')
    result = yield ap.del_client(wlan_id, mac, ban_time=1000)
    raise_on_ubus_error(result, 'cannot deauth client')
    raise Return((True, '%s banned' % mac))


@coroutine
def client_permit(ap, wlan_id, mac):
    result = yield ap.set_mac_filter_policy('deny', wlan_id)
    raise_on_ubus_error(result, 'cannot set policy')
    mac_list = yield ap.get_mac_list(wlan_id)
    if mac not in mac_list:
        raise Return((False, "mac doesn't exists"))
    mac_list.remove(mac)
    result = yield ap.set_mac_list(list(set(mac_list)), wlan_id)
    raise_on_ubus_error(result, 'cannot set maclist')
    result = yield ap.commit('wireless')
    raise_on_ubus_error(result, 'cannot commit configuration')
    raise Return((True, '%s permitted' % mac))


@coroutine
def ap_list(ap_list):
    ap_all = []
    for host, ap in six.iteritems(ap_list):
        status = dict(host=host)
        system_status = yield ap.get_system_status()
        status['system'] = system_status
        board_status = yield ap.get_board_status()
        status['board'] = board_status
        hostapd_clients = yield ap.list_clients()
        num_clients = sum([len(c[1]['clients'])
                           for c in six.iteritems(hostapd_clients)])
        status['num_clients'] = num_clients
        ap_all.append(status)

    raise Return(ap_all)


@coroutine
def wlan_list(ap_list):
    ap_all = []
    for host, ap in six.iteritems(ap_list):
        status = dict(host=host)
        board_status = yield ap.get_board_status()
        status['board'] = board_status
        wireless_status = yield ap.get_wireless_status()
        status['wireless'] = wireless_status
        ap_all.append(status)

    raise Return(ap_all)


@coroutine
def wlan_set_channel(ap, channel, wlan_id):
    result = yield ap.set_channel(channel, wlan_id)
    raise_on_ubus_error(result)
    result = yield ap.commit('wireless')
    raise_on_ubus_error(result, 'cannot commit configuration')
    raise Return((True, 'channel changed to %d' % channel))


@coroutine
def wlan_set_ssid(ap, ssid, wlan_id):
    result = yield ap.set_ssid(ssid, wlan_id)
    raise_on_ubus_error(result)
    result = yield ap.commit('wireless')
    raise_on_ubus_error(result, 'cannot commit configuration')
    raise Return((True, 'ssid changed to %s' % ssid))


@coroutine
def wlan_set_encryption(ap, method, key, wlan_id):
    result = yield ap.set_encryption(method, key, wlan_id)
    raise_on_ubus_error(result)
    result = yield ap.commit('wireless')
    raise_on_ubus_error(result, 'cannot commit configuration')
    raise Return((True, 'encryption changed to %s(%s)' % (key, method)))


@coroutine
def wlan_enable(ap, enable, wlan_id):
    result = yield ap.enable_wlan(enable, wlan_id)
    raise_on_ubus_error(result)
    result = yield ap.commit('wireless')
    raise_on_ubus_error(result, 'cannot commit configuration')
    raise Return((True, '%s' % ('enabled' if enable else 'disabled')))



#
# This piece of code is written by
#    Jianing Yang <jianingy.yang@gmail.com>
# with love and passion!
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#                              2 Mar, 2016
#


class GenericError(Exception):
    error_format = 'generic error'

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if not message:
            try:
                message = self.error_format % self.kwargs
            except:
                message = 'cannot format exception message'

        self.message = message
        super(GenericError, self).__init__(message)



#
# This piece of code is written by
#    Jianing Yang <jianingy.yang@gmail.com>
# with love and passion!
#
#        H A P P Y    H A C K I N G !
#              _____               ______
#     ____====  ]OO|_n_n__][.      |    |
#    [________]_|__|________)<     |YANG|
#     oo    oo  'oo OOOO-| oo\\_   ~o~~o~
# +--+--+--+--+--+--+--+--+--+--+--+--+--+
#                             18 Mar, 2016
#
from click import echo, style
import sys


def debug(msg):
    echo(style('[#] ' + msg, fg='white'))


def info(msg):
    echo(style('[*] ' + msg, fg='blue'))


def success(msg):
    echo(style('[+] ' + msg, fg='green'))


def error(msg):
    echo(style('[!] ' + msg, fg='red'))


def fatal(msg, code=111):
    echo(style('[!] ' + msg, fg='red'))
    sys.exit(code)


def out(msg):
    echo(msg)


def header_out(msg):
    echo(style(msg, fg='blue'), err=True)
